;;;;
;;;; gaf-read -- Macro that generates gaf-read based on the extant gaf
;;;; version specification.  gaf-read turns a gschem schematic or
;;;; symbol file into an S-expression.
;;;;
;;;; Copyright 2007 Dean Ferreyra, All rights reserved
;;;;
;;;; This file is part of octw-guile-geda.
;;;;
;;;; octw-guile-geda is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; octw-guile-geda is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with octw-guile-geda.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;; $Id: gaf-read.scm,v 1.1 2008-09-04 03:40:47 dean Exp $

(define-module (octw gaf-read))

(use-modules (srfi srfi-1)
	     (srfi srfi-13)
	     (srfi srfi-19))
(use-modules (ice-9 optargs)
             (ice-9 rdelim))
(use-modules (octw gaf))
(use-syntax (ice-9 syncase))

(export gaf-read-line
        gaf-read-gen)

(define (gaf-read-line in)
  (let ((line (read-line in 'concat)))
    (if (eof-object? line)
        line
        (string-drop-right line 1))))

;;;
;;; gaf-read-gen
;;;
;;; This macro generates the function gaf-read using the extant gaf
;;; version specification.  Here's an example of how to use it at the
;;; top level:
;;;
;;;   ;;; Required modules.
;;;   (use-modules (octw conditional)
;;;   	           (octw anaphoric)
;;;   	           (octw gaf)
;;;                (octw gaf-v1)
;;;                (octw gaf-read))
;;; 
;;;   ;;; Generate gEDA/gaf version 1 objects.
;;;   (gaf-gen-v1-specs)
;;;   ;;; Generate version 1 reader.
;;;   (gaf-read-gen)
;;;
;;; This is a macro so that functions generated by the gaf-v1-spec
;;; macro are available.
(define-syntax gaf-read-gen
  (lambda (form)
    (syntax-case form ()
      ((_)
       (syntax
        (gaf-read-gen gaf-fields->object)))
      ((_ fields->object)
       (syntax
        (define* (gaf-read in #:optional (sexp '()))
          (let ((line (gaf-read-line in)))
            (if (eof-object? line)
                (reverse! sexp)
                (let ((fields (string-split line #\space)))
                  (let ((obj (fields->object fields)))
                    (if (not obj)
                        (throw 'gaf-unrecognized-line
                               (format #f "Unrecognized line: ~S" line))
                        (case (object-type obj)
                          ((embedded-start)
                           ;; Collect embedded component objects
                           (let ((subs (gaf-read in '()))
                                 (obj (first sexp)))
                             (unless (eq? (object-type obj) 'component)
                               (throw 'gaf-expected-component
                                      (format #f
                                              (string-append
                                               "Expected component "
                                               "while processing "
                                               "embedded objects, but got: ~S")
                                              obj)))
                             (let ((subs-obj (cons 'embedded subs)))
                               (if (gaf-composite? obj)
                                   (set-car! sexp (append obj (list subs-obj)))
                                   (set-car! sexp (list obj subs-obj))))
                             (gaf-read in sexp)))
                          ((attributes-start)
                           ;; Collect object attributes
                           (let ((subs (gaf-read in '()))
                                 (obj (first sexp)))
                             (let ((subs-obj (cons 'attributes subs)))
                               (if (gaf-composite? obj)
                                   (set-car! sexp (append obj (list subs-obj)))
                                   (set-car! sexp (list obj subs-obj))))
                             (gaf-read in sexp)))
                          ((embedded-end attributes-end)
                           ;; Done with embedded or attributes section
                           (reverse! sexp))
                          ((text)
                           ;; Need to read the right number of lines
                           (let collect-text ((i (text-num_lines obj))
                                              (lines '()))
                             (if (= i 0)
                                 (gaf-read in
                                           (cons
                                            (append obj
                                                    (list (reverse! lines)))
                                            sexp))
                                 (let ((line (gaf-read-line in)))
                                   (collect-text (- i 1)
                                                 (cons line lines))))))
                          (else
                           (gaf-read in (cons obj sexp)))))))))))))))
