;;;;
;;;; gaf-write -- Macro that generates gaf-write based on the extant
;;;; gaf version specification.  gaf-write turns an S-expression into
;;;; a gschem schematic or symbol file.
;;;;
;;;; Copyright 2007 Dean Ferreyra, All rights reserved
;;;;
;;;; This file is part of octw-guile-geda.
;;;;
;;;; octw-guile-geda is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; octw-guile-geda is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with octw-guile-geda.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;; $Id: gaf-write.scm,v 1.1 2008-09-04 03:40:47 dean Exp $

(define-module (octw gaf-write))

(use-modules (srfi srfi-1)
	     (srfi srfi-13)
	     (srfi srfi-19))
(use-modules (ice-9 optargs)
             (ice-9 rdelim))
(use-modules (octw gaf))
(use-syntax (ice-9 syncase))

(export gaf-write-gen)

;;;
;;; gaf-write-gen
;;;
;;; This macro generates the function gaf-write using the extant gaf
;;; version specification.  Here's an example of how to use it at the
;;; top level:
;;;
;;;   ;;; Required modules.
;;;   (use-modules (octw conditional)
;;;   	           (octw anaphoric)
;;;   	           (octw gaf)
;;;                (octw gaf-v1)
;;;                (octw gaf-write))
;;; 
;;;   ;;; Generate gEDA/gaf version 1 objects.
;;;   (gaf-gen-v1-specs)
;;;   ;;; Generate version 1 writer.
;;;   (gaf-write-gen)
;;;
;;; This is a macro so that functions generated by the gaf-v1-spec
;;; macro are available.
(define-syntax gaf-write-gen
  (lambda (form)
    (syntax-case form ()
      ((_)
       (syntax
        (gaf-write-gen gaf-object->fields)))
      ((_ object->fields)
       (syntax
        (define (gaf-write out sexp)
          (for-each
           (lambda (obj)
             (let ((fields (object->fields obj)))
               (if (gaf-composite? obj)
                   (gaf-write out obj)
                   (case (object-type obj)
                     ((embedded)
                      (display #\[ out)
                      (newline out)
                      (gaf-write out (cdr obj))
                      (display #\] out)
                      (newline out))
                     ((attributes)
                      (display #\{ out)
                      (newline out)
                      (gaf-write out (cdr obj))
                      (display #\} out)
                      (newline out))
                     (else
                      (let ((fields (object->fields obj)))
                        (if (not fields)
                            (announce-error
                             "Could not convert object into fields: ~S" obj)
                            (begin
                              (display (gaf-fields->line fields) out)
                              (newline out))))
                      (case (object-type obj)
                        ((text)
                         ;; Handle extra text arguments
                         (for-each (lambda (line)
                                     (display line out)
                                     (newline out))
                                   (text-lines obj)))))))))
           sexp)))))))
