;;;;
;;;; gaf
;;;;
;;;; Copyright 2006 Dean Ferreyra, All rights reserved
;;;;
;;;; This file is part of octw-guile-geda.
;;;;
;;;; octw-guile-geda is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; octw-guile-geda is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with octw-guile-geda.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;; $Id: gaf.scm,v 1.2 2008-09-08 18:54:52 dean Exp $

(define-module (octw gaf))

(export gaf-spec-name
	gaf-spec-type
	gaf-spec-fields
	gaf-spec-type-equal?
	gaf-spec-field-name
	gaf-spec-field-type
	any->type
	string->type
        type->string
	validate-type
        gaf-composite?
        gaf-composite-attributes
        gaf-base-text-attribute?
        gaf-text-key-value
        gaf-base-composite-attribute-text-obj
        gaf-base-composite-attribute-value
        gaf-base-composite-attribute-value-set!
        gaf-base-composite-attribute-add!
        gaf-fields->line
	object-type
	object-fields
        make-object
        object-field
        object-field-set!
        gaf-gen
        gaf-gen-specs
        gen-spec-name->spec
        gen-object->fields
        gen-makers-setters-getters
        gen-spec-maker-setter-getter
        gen-field-spec-setter-getter
        gen-spec-maker
        gen-field-spec-getter
        gen-field-spec-setter
        gen-type-field->maker
        gen-fields->object
	)

(use-modules (srfi srfi-1)
             (srfi srfi-11)
             (srfi srfi-13))
(use-modules (ice-9 rdelim))
(use-modules (octw conditional)
             (octw anaphoric))

(use-syntax (ice-9 syncase))

;;;
;;; Specs are in the form '(spec-name spec-character (spec-fields ...))
;;;

(define (gaf-spec-name spec)
  (first spec))

(define (gaf-spec-type spec)
  (second spec))

(define (gaf-spec-fields spec)
  (third spec))

(define (gaf-spec-type-equal? spec str)
  (and (= (string-length str) 1)
       (eq? (string-ref str 0)
	    (gaf-spec-type spec))))

;;;
;;; Field specs are in the form '(field-name field-type) where
;;; field-type is the symbol int, char, or string.
;;;

(define (gaf-spec-field-name field-spec)
  (first field-spec))

(define (gaf-spec-field-type field-spec)
  (second field-spec))

;;;
;;; Type functions
;;;

(define (any->type field-spec val)
  (if (string? val)
      (string->type field-spec val)
      (validate-type field-spec val)))

(define (string->type field-spec str)
  (let ((val (case (gaf-spec-field-type field-spec)
	       ((char)
		(string-ref str 0))
	       ((int)
		(string->number str))
	       (else
		str))))
    (validate-type field-spec val)))

(define (type->string field-spec val)
  (validate-type field-spec val)
  (case (gaf-spec-field-type field-spec)
    ((char)
     (string val))
    ((int)
     (number->string val))
    (else
     val)))

(define (validate-type field-spec val)
  (case (gaf-spec-field-type field-spec)
    ((char)
     (unless (char? val)
       (throw 'gaf-spec-type-error "Expected a character but got ~S" val)))
    ((int)
     (unless (integer? val)
       (throw 'gaf-spec-type-error "Expected an integer but got ~S" val)))
    ((string)
     (unless (string? val)
       (throw 'gaf-spec-type-error "Expected a string but got ~S" val))))
  val)

(define (gaf-composite? obj)
  (and (pair? obj)
       (pair? (first obj))))

(define (gaf-composite-attributes obj)
  (find (lambda (obj)
          (eq? (object-type obj) 'attributes))
        obj))

(define (find-map pred lst)
  (let loop ((lst lst))
    (if (null? lst)
        #f
        (aif (pred (car lst))
             it
             (loop (cdr lst))))))

(define (gaf-base-text-attribute? text-lines-fn obj)
  (and (eq? (object-type obj) 'text)
       (string-index (first (text-lines-fn obj)) #\=)
       #t))

(define (gaf-text-key-value text)
  (let ((equal-index (string-index text #\=)))
    (let ((key (string-take text equal-index))
          (val (string-drop text (+ equal-index 1))))
      (values key val))))

(define (gaf-base-composite-attribute-text-obj text-lines-fn obj key)
  (aand (gaf-composite-attributes obj)
        (find-map
         (lambda (obj)
           (and (gaf-base-text-attribute? text-lines-fn obj)
                (let-values (((text-key text-val)
                              (gaf-text-key-value
                               (first (text-lines-fn obj)))))
                  (and (string=? key text-key)
                       obj))))
         (cdr it))))

(define (gaf-base-composite-attribute-value text-lines-fn obj key)
  (aand (gaf-composite-attributes obj)
        (find-map
         (lambda (obj)
           (and (gaf-base-text-attribute? text-lines-fn obj)
                (let-values (((text-key text-val)
                              (gaf-text-key-value
                               (first (text-lines-fn obj)))))
                  (and (string=? key text-key)
                       text-val))))
         (cdr it))))

(define (gaf-base-composite-attribute-value-set! text-lines-fn obj key val)
  (aand (gaf-composite-attributes obj)
        (find-map
         (lambda (obj)
           (and (gaf-base-text-attribute? text-lines-fn obj)
                (let-values (((text-key text-val)
                              (gaf-text-key-value
                               (first (text-lines-fn obj)))))
                  (and (string=? key text-key)
                       (begin
                         (set-car! (text-lines-fn obj)
                                   (format #f "~A=~A" key val))
                         text-val)))))
         (cdr it))))

(define (gaf-base-composite-attribute-add! text-lines-set!-fn
                                           obj key val text-obj-template)
  (let ((text-obj (list-copy text-obj-template))
        (last-attrib (last-pair (gaf-composite-attributes obj))))
    (text-lines-set!-fn text-obj
                        (list (format #f "~A=~A" key val)))
    (set-cdr! last-attrib (list text-obj))))

(define (gaf-fields->line fields)
  (string-join fields " "))

;;; object-type
;;;
;;; Given an object, return its type
;;;
(define (object-type obj)
  (first obj))

;;; object-fields
;;;
;;; Given an object, return its fields
;;;
(define (object-fields obj)
  (cdr obj))

;;;
;;; Helper functions used by spec macros
;;;

(define (make-object spec fields)
  (unless (gaf-spec-type-equal? spec (first fields))
    (throw 'gaf-make-object-error
           (format #f
                   (string-append
                    "While processing ~S object, "
                    "expected type field, \"~A\", "
                    "but got: \"~A\"")
                   (gaf-spec-name spec)
                   (gaf-spec-type spec)
                   (first fields))))
  (let ((field-specs (gaf-spec-fields spec)))
    (let loop ((field-specs field-specs)
	       (fields (cdr fields))
	       (obj `(,(gaf-spec-name spec))))
      (if (null? field-specs)
	  (begin
	    (unless (null? fields)
	      (throw 'gaf-make-object-error
                     (format #f
                             "While procesing ~S object, got too many fields"
                             (gaf-spec-name spec))))
 	    (reverse! obj))
	  (begin
	    (when (null? fields)
	      (throw 'gaf-make-object-error
                     (format #f
                             "While processing ~S object, got too few fields"
                             (gaf-spec-name spec))))
	    (loop (cdr field-specs) (cdr fields)
		  (cons (string->type (car field-specs) (car fields))
			obj)))))))

(define (object-field obj field-index)
  (list-ref obj field-index))

(define (object-field-set! obj field-spec field-index val)
  (set-car! (drop obj field-index)
	    (any->type field-spec val))
  obj)

;;;
;;; Macros to generate creators, setters, and getters from a list of
;;; specs.
;;;

;;; gaf-gen -- Given a with-gaf-*-specs macro, defines all the makers,
;;; setters, getters, and some useful top-level functions:
;;;
;;;   (gaf-gen with-gaf-v1-specs)
;;;
(define-syntax gaf-gen
  (lambda (form)
    (syntax-case form ()
      ((_ with-specs-macro)
       (syntax
        (begin
          (define *gaf-specs*
            (with-specs-macro gaf-gen-specs))
          (define gaf-spec-name->spec
            (with-specs-macro gen-spec-name->spec))
          (define gaf-object->fields
            (with-specs-macro gen-object->fields gaf-spec-name->spec))
          (define gaf-type-field->maker
            (with-specs-macro gen-type-field->maker))
          (define gaf-fields->object
            (with-specs-macro gen-fields->object gaf-type-field->maker))
          (with-specs-macro gen-makers-setters-getters)))))))

;;; gaf-gen-specs -- Yields the gaf specs as a quoted list
;;;
(define-syntax gaf-gen-specs
  (lambda (form)
    (syntax-case form ()
      ((_ specs)
       (syntax 'specs)))))

;;; gen-spec-name->spec -- Yields a function that takes the spec name;
;;; e.g., 'version, and evaluates to the spec; e.g., '(version #\v
;;; ((version int) (fileformat_version in)))
;;;
(define-syntax gen-spec-name->spec
  (lambda (form)
    (syntax-case form ()
      ((_ (spec))
       (with-syntax (((spec-name spec-type spec-fields) (syntax spec)))
         (syntax
          (lambda (name)
            (and (eq? 'spec-name name) 'spec)))))
      ((_ (spec rest ...))
       (with-syntax (((spec-name spec-type spec-fields) (syntax spec)))
         (syntax
          (lambda (name)
            (if (eq? 'spec-name name)
                'spec
                ((gen-spec-name->spec (rest ...)) name)))))))))

;;; gen-object->fields -- Yields a function that takes an object and
;;; evaluates to a list of fields suitable for string-joining and
;;; outputting to a file; e.g., '("v" "20" "30").
;;;
(define-syntax gen-object->fields
  (lambda (form)
    (syntax-case form ()
      ((_ specs spec-name->spec)
       (syntax
        (lambda (obj)
          (let ((spec (gaf-spec-name->spec (object-type obj))))
            (and spec
                 (let ((type-field (gaf-spec-type spec))
                       (field-specs (gaf-spec-fields spec)))
                   (append (list (string type-field))
                           (map (lambda (spec val)
                                  (type->string spec val))
                                field-specs
                                (object-fields obj))))))))))))

;;; gen-makers-setters-getters -- Defines functions for the object
;;; makers and field setters and getters.
(define-syntax gen-makers-setters-getters
  (lambda (form)
    (syntax-case form ()
      ((_ (spec))
       (syntax
        (gen-spec-maker-setter-getter spec)))
      ((_ (spec rest ...))
       (syntax
        (begin
          (gen-spec-maker-setter-getter spec)
          (gen-makers-setters-getters (rest ...))))))))

(define-syntax gen-spec-maker-setter-getter
  (lambda (form)
    (syntax-case form ()
      ((_ (spec-name spec-type ()))
       (syntax
        (begin
          (gen-spec-maker (spec-name spec-type ())))))
      ((_ (spec-name spec-type spec-fields))
       (syntax
        (begin
          (gen-spec-maker (spec-name spec-type spec-fields))
          (gen-field-spec-setter-getter (spec-name spec-type spec-fields)
                                        spec-fields 1)))))))

(define-syntax gen-field-spec-setter-getter
  (lambda (form)
    (syntax-case form ()
      ((_ spec (field-spec) index)
       (syntax
	(begin
	  (gen-field-spec-getter spec field-spec index)
	  (gen-field-spec-setter spec field-spec index))))
      ((_ spec (field-spec rest ...) index)
       (with-syntax ((next-index (datum->syntax-object
				  (syntax index)
				  (+ 1 (syntax-object->datum
					(syntax index))))))
	 (syntax
	  (begin
	    (gen-field-spec-getter spec field-spec index)
	    (gen-field-spec-setter spec field-spec index)
	    (gen-field-spec-setter-getter spec (rest ...) next-index ))))))))

(define-syntax gen-spec-maker
  (lambda (form)
    (syntax-case form ()
      ((_ spec)
       (with-syntax (((spec-name spec-type spec-fields) (syntax spec)))
         (with-syntax ((make-fn-name
                        (datum->syntax-object
                         (syntax spec)
                         (symbol-append
                          'make-
                          (syntax-object->datum (syntax spec-name))))))
           (syntax
            (begin
              (define make-fn-name
                (let ((spec-e 'spec))
                  (lambda (fields)
                    (make-object spec-e fields))))))))))))

(define-syntax gen-field-spec-getter
  (lambda (form)
    (syntax-case form ()
      ((_ spec field-spec field-index)
       (with-syntax ((getter-fn-name
		      (datum->syntax-object
		       (syntax field-spec)
		       (symbol-append
			(gaf-spec-name (syntax-object->datum (syntax spec)))
			'-
			(gaf-spec-field-name
			 (syntax-object->datum (syntax field-spec)))))))
	 (syntax
	  (begin
	    (define (getter-fn-name obj)
	      (object-field obj field-index)))))))))

(define-syntax gen-field-spec-setter
  (lambda (form)
    (syntax-case form ()
      ((_ spec field-spec field-index)
       (with-syntax ((setter-fn-name
		      (datum->syntax-object
		       (syntax field-spec)
		       (symbol-append
			(gaf-spec-name (syntax-object->datum (syntax spec)))
			'-
			(gaf-spec-field-name
			 (syntax-object->datum (syntax field-spec)))
			'-set!))))
	 (syntax
	  (begin
	    (define setter-fn-name
	      (let ((field-spec-e 'field-spec))
		(lambda (obj val)
		  (object-field-set! obj
				     field-spec-e field-index
				     val)))))))))))

;;; gen-type-field->maker -- Yields a function that takes a type-field;
;;; e.g., #\v, and evaluates to the maker for that object; e.g.,
;;; make-version.
;;;
(define-syntax gen-type-field->maker
  (lambda (form)
    (syntax-case form ()
      ((_ (spec))
       (with-syntax ((make-fn-name
                      (datum->syntax-object
                       (syntax spec)
                       (symbol-append
                        'make-
                        (gaf-spec-name
                         (syntax-object->datum (syntax spec)))))))
         (syntax
          (lambda (type-field)
            (and (gaf-spec-type-equal? 'spec type-field)
                 make-fn-name)))))
      ((_ (spec rest ...))
       (with-syntax ((make-fn-name (datum->syntax-object
                                    (syntax spec)
                                    (symbol-append
                                     'make-
                                     (gaf-spec-name
                                      (syntax-object->datum (syntax spec)))))))
         (syntax
          (lambda (type-field)
            (if (gaf-spec-type-equal? 'spec type-field)
                make-fn-name
                ((gen-type-field->maker (rest ...)) type-field)))))))))

;;; gen-fields->object -- Given the specs and the type-field->maker
;;; function, yields a function that converts a list of fields into the
;;; appropriate object.
;;;
(define-syntax gen-fields->object
  (lambda (form)
    (syntax-case form ()
      ((_ specs type-field->maker)
       (syntax
	(lambda (fields)
	  (let ((maker (type-field->maker (first fields))))
	    (and maker (maker fields)))))))))
