;;;;
;;;; pads -- Defines functions and macros for generating PCB
;;;; footprints programmatically.
;;;;
;;;; Copyright 2007 Dean Ferreyra, All rights reserved
;;;;
;;;; This file is part of octw-guile-geda.
;;;;
;;;; octw-guile-geda is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; octw-guile-geda is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with octw-guile-geda.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;; $Id: pads.scm,v 1.2 2008-09-08 18:53:40 dean Exp $

(define-module (octw pads))

(export mm->in
        in->mils
        mils->centimils
        mm->mils
        mm->centimils
        in->centimils
        convert-mm->centimils
        combine
        let-deep-clone
        move-dx
        move-dy
        move
        rotate
        mirror-x
        mirror-y
        move-along
        move-along-x
        move-along-y
        indent
        output
        <pt>
        x
        y
        make-pt
        distance-x
        distance-y
        distance
        angle->unit-vector
        *origin*
        *along-x*
        *along-y*
        <stroke>
        center
        place
        stroke-length
        stroke-width
        grow
        <flagable>
        <nameable>
        <numberable>
        <pad>
        start-pt
        end-pt
        make-pad
        make-pad-x
        make-pad-y
        mask-length
        mask-width
        clearance-length
        clearance-width
        <element-line>
        make-element-line
        make-element-line-x
        make-element-line-y
        <element-arc>
        make-element-arc
        make-element-circle
        <pin>
        <via>
        <element>
        outline
        make-outline
        remove-flag
        remove-flags
        add-flag
        replace-flag
        shape-round
        shape-round-first
        shape-round-last
        shape-square
        shape-square-first
        shape-square-last
        name
        name*
        number
        enumerate
        enumerate*
        duplicate
        row
        make-polar-outline-top
        make-polar-outline-right
        make-polar-outline-bottom
        make-polar-outline-left
        make-scant-outline
        make-scant-polar-outline-top
        make-scant-polar-outline-right
        make-scant-polar-outline-bottom
        make-scant-polar-outline-left
        make-notched-outline-ul
        make-notched-outline-ul
        make-notched-outline-ur
        make-notched-outline-lr
        make-notched-outline-ll
        make-notched-outline-top
        make-notched-outline-top
        make-notched-outline-right
        make-notched-outline-bottom
        make-notched-outline-left
        output-to-file)

(use-modules (oop goops))
(use-modules (srfi srfi-1)
	     (srfi srfi-9)
	     (srfi srfi-13))
(use-modules (ice-9 optargs)
             (ice-9 rdelim))
(use-modules (octw conditional)
             (octw anaphoric))
(use-syntax (ice-9 syncase))

(define-syntax let-deep-clone
  (lambda (form)
    (syntax-case form ()
      ((_ ((cloned-obj obj) ...)
          body ...)
       (syntax
        (let ((cloned-obj (deep-clone obj)) ...)
          body ...))))))

(define-syntax strip-types
  (lambda (form)
    (syntax-case form ()
      ((_ (call-expr ...))
       (with-syntax (((stripped-call-expr ...)
                      (map (lambda (elem-syn)
                             (syntax-case elem-syn ()
                               ((arg type) (syntax arg))
                               (arg (syntax arg))))
                           (syntax (call-expr ...)))))
         (syntax
          (stripped-call-expr ...)))))))

(define-syntax define-list-method
  (lambda (form)
    (syntax-case form ()
      ((_ (meth args ...))
       (syntax
        (define-method (meth (lst <list>) args ...)
          (map (lambda (elem)
                 (strip-types (meth elem args ...)))
               lst)))))))

(define-syntax define-rotate-methods
  (lambda (form)
    (syntax-case form ()
      ((_ base-fn ((fn angle) ...))
       (syntax
        (begin
          (define-method (fn dx dy . rest)
            (if (or (= angle 90) (= angle -90))
                (rotate (apply base-fn dy dx rest) angle)
                (rotate (apply base-fn dx dy rest) angle)))
          ...))))))

(define-syntax define-slots-op
  (lambda (form)
    (syntax-case form ()
      ((_ (meth (obj obj-type) args ...) (slot-sym ...))
       (syntax
        (define-method (meth (obj obj-type) args ...)
          (let-deep-clone ((new-obj obj))
            (slot-mod! new-obj 'slot-sym
                       (strip-types (meth <> args ...))) ...
            new-obj)))))))

(define-syntax define-slots-ops
  (lambda (form)
    (syntax-case form ()
      ((_ (function-signature ...)
          (slot-sym ...))
       (syntax
        (begin
          (define-slots-op function-signature (slot-sym ...)) ...))))))

(define PI (* 4 (atan 1 1)))

(define (deg->rad degrees)
  (* 2 PI (/ degrees 360)))

(define (sq x)
  (* x x))

(define (mm->in x)
  (/ x 25.4))

(define (in->mm x)
  (* x 25.4))

(define (in->mils x)
  (* x 1000))

(define (mils->centimils x)
  (* x 100))

(define mm->mils (compose in->mils mm->in))

(define mm->centimils (compose mils->centimils in->mils mm->in))

(define in->centimils (compose mils->centimils in->mils))

(define-method (convert-mm->centimils x)
  (mm->centimils x))

;;; combine
;;;
;;; Combine elements into a flat list.
;;;
(define (combine . rest)
  (if (= (length rest) 1)
      rest
      (apply append (map (lambda (elem)
                           (if (pair? elem) elem (list elem)))
                         rest))))

(define (modify-element fn lst index)
  (let loop ((lst lst)
             (i 0)
             (acc '()))
    (if (null? lst)
        (reverse! acc)
        (loop (cdr lst) (+ i 1)
              (cons (if (= i index)
                        (fn (car lst))
                        (car lst))
                    acc)))))
            
(define (move-dx obj dx)
  (move obj dx 0))

(define (move-dy obj dy)
  (move obj 0 dy))

(define-method (deep-clone (lst <list>))
  (map deep-clone lst))

(define-list-method (move dx dy))
(define-list-method (rotate degrees))
(define-list-method (mirror-x))
(define-list-method (mirror-y))
(define-list-method (convert-mm->centimils))

(define (indent level)
  (make-string (* 4 level) #\space))

(define-method (output obj out (level <integer>))
  (format out "~A~A" (indent level) obj))

(define-method (output obj out)
  (output obj out 0))

(define-method (output (lst <list>) out level)
  (format out "~A"
          (string-join
           (map (lambda (elem) (output elem #f level)) lst)
           (string #\newline))))

;;;
;;; pt
;;;

(define-class <pt> ()
  (x #:accessor x
     #:init-keyword #:x
     #:init-value 0)
  (y #:accessor y
     #:init-keyword #:y
     #:init-value 0))

(define (make-pt x y)
  (make <pt> #:x x #:y y))

(define *origin* (make-pt 0 0))
(define *along-x* (make-pt 1 0))
(define *along-y* (make-pt 0 1))

(define-method (object-equal? (a <pt>) (b <pt>))
  (and (= (x a) (x b))
       (= (y a) (y b))))

(define-method (distance-x (a <pt>) (b <pt>))
  (- (x a) (x b)))

(define-method (distance-y (a <pt>) (b <pt>))
  (- (y a) (y b)))

(define-method (distance (a <pt>) (b <pt>))
  (sqrt (+ (sq (distance-x a b))
           (sq (distance-y a b)))))

(define-method (+ (a <pt>) (b <pt>))
  (make-pt (+ (x a) (x b))
           (+ (y a) (y b))))

(define-method (/ (pt <pt>) (n <number>))
  (make-pt (/ (x pt) n)
           (/ (y pt) n)))

(define-method (- (pt <pt>))
  (make-pt (- (x pt)) (- (y pt))))

(define-method (- (a <pt>) (b <pt>))
  (make-pt (distance-x a b) (distance-y a b)))

(define (radians->unit-vector theta)
  (make-pt (cos theta) (sin theta)))

(define (angle->unit-vector degrees)
  (radians->unit-vector (deg->rad degrees)))

;;; Create a unit vector with the same angle as the vector from
;;; START-PT to END-PT.
(define-method (unit-vector (start-pt <pt>) (end-pt <pt>))
  (let ((theta (atan (distance-y end-pt start-pt)
                     (distance-x end-pt start-pt))))
    (radians->unit-vector theta)))

;;; Create a unit vector with the same angle as the vector from
;;; the origin to END-PT.
(define-method (unit-vector (end-pt <pt>))
  (unit-vector *origin* end-pt))

(define-method (move (pt <pt>) dx dy)
  (make-pt (+ (x pt) dx) (+ (y pt) dy)))

;;; Rotate point about the origin.
(define-method (rotate (pt <pt>) degrees)
  ;; Reverse rotation direction so that positive angles yield the
  ;; normal CCW rotation in PCB.
  (let ((degrees (- degrees)))
    (let ((theta (deg->rad degrees)))
      (make-pt (- (* (x pt) (cos theta))
                  (* (y pt) (sin theta)))
               (+ (* (x pt) (sin theta))
                  (* (y pt) (cos theta)))))))

(define-method (mirror-x (pt <pt>))
  (make-pt (x pt) (- (y pt))))

(define-method (mirror-y (pt <pt>))
  (make-pt (- (x pt)) (y pt)))

(define-method (move-along (pt <pt>) distance (direction-unit-vect <pt>))
  (make-pt (+ (x pt) (* distance (x direction-unit-vect)))
           (+ (y pt) (* distance (y direction-unit-vect)))))

(define-list-method (move-along distance (distance-unit-vect <pt>)))

(define-method (move-along-x obj distance)
  (move-along obj distance *along-x*))

(define-method (move-along-y obj distance)
  (move-along obj distance *along-y*))

(define-list-method (move-along-x distance))
(define-list-method (move-along-y distance))

(define-method (move obj (pt <pt>))
  (move obj (x pt) (y pt)))

(define-method (convert-mm->centimils (pt <pt>))
  (make-pt (convert-mm->centimils (x pt))
           (convert-mm->centimils (y pt))))

(define-method (center (pt <pt>))
  pt)

(define-method (center (objs <list>))
  (let ((n (length objs))
        (sum (fold (lambda (obj acc)
                     (+ (center obj) acc))
                     *origin* objs)))
    (/ sum n)))

(define-method (output (pt <pt>) out level)
  (format out "~A ~A"
          (to-exact (x pt))
          (to-exact (y pt))))

(define-method (place obj (center-pt <pt>))
  (move obj (- center-pt (center obj))))

;;;
;;; stroke
;;;

(define-class <stroke> ()
  (start-pt #:accessor start-pt
            #:init-keyword #:start-pt
            #:init-value (make-pt 0 0))
  (end-pt #:accessor end-pt
          #:init-keyword #:end-pt
          #:init-value (make-pt 0 0))
  (width #:accessor width
         #:init-keyword #:width
         #:init-value 0))

(define (make-stroke start-pt end-pt width)
  (make <stroke> #:start-pt start-pt #:end-pt end-pt #:width width))

(define-slots-ops ((move (obj <stroke>) dx dy)
                   (move-along (obj <stroke>) distance
                               (direction-unit-vect <pt>))
                   (rotate (obj <stroke>) degrees)
                   (mirror-x (obj <stroke>))
                   (mirror-y (obj <stroke>)))
  (start-pt
   end-pt))

(define-slots-op (convert-mm->centimils (obj <stroke>))
  (start-pt
   end-pt
   width))

(define-method (stroke-length (stroke <stroke>))
  (distance (start-pt stroke) (end-pt stroke)))

(define-method (stroke-width (stroke <stroke>))
  (slot-ref stroke 'width))

;;; Create a new stroke where the length is the old width and the
;;; width is the old length.
(define-method (swap-axis (stroke <stroke>))
  (with-slots stroke ((start-pt 'start-pt)
                      (end-pt 'end-pt)
                      (width 'width))
    (let ((width (stroke-length stroke))
          (half-length (/ width 2))
          (center-pt (center stroke))
          (direction (rotate (unit-vector start-pt end-pt) 90)))
      (let ((start-pt (move-along center-pt (- half-length) direction))
            (end-pt (move-along center-pt half-length direction)))
        (let-deep-clone ((stroke stroke))
          (slot-set! stroke 'start-pt start-pt)
          (slot-set! stroke 'end-pt end-pt)
          (slot-set! stroke 'width width)
          stroke)))))

(define-method (grow (stroke <stroke>) in-length in-width)
  (let-deep-clone ((stroke stroke))
    (slot-mod! stroke 'width (+ <> in-width))
    (with-slots stroke ((start-pt 'start-pt)
                        (end-pt 'end-pt))
      (let ((direction (unit-vector start-pt end-pt)))
        (slot-mod! stroke 'start-pt
                   (move-along <> (/ in-length 2) (- direction)))
        (slot-mod! stroke 'end-pt
                   (move-along <> (/ in-length 2) direction))))
    stroke))

(define-method (center (stroke <stroke>))
  (move (start-pt stroke)
        (/ (distance-x (end-pt stroke) (start-pt stroke)) 2)
        (/ (distance-y (end-pt stroke) (start-pt stroke)) 2)))

(define-class <flagable> ()
  (flags #:init-keyword #:flags #:init-value ""))

(define-class <nameable> ()
  (name #:init-keyword #:name
        #:init-value ""))

(define-class <numberable> ()
  (number #:init-keyword #:number
          #:init-value ""))

;;;
;;; pad
;;;

(define-class <pad> (<flagable> <nameable> <numberable> <stroke>)
  (mask-extent #:init-keyword #:mask-extent
               #:init-value 0)
  (clearance-extent #:init-keyword #:clearance-extent
                    #:init-value 0)
  (flags #:init-value "square"))

;;; Make a pad at the origin.
(define-method (make-pad length width (direction-unit-vect <pt>) . rest)
  (let ((half-length (/ length 2)))
    (let ((start-pt (make-pt (- (* (x direction-unit-vect) half-length))
                             (- (* (y direction-unit-vect) half-length))))
          (end-pt (make-pt (+ (* (x direction-unit-vect) half-length))
                           (+ (* (y direction-unit-vect) half-length)))))
      (apply make <pad>
             #:start-pt start-pt
             #:end-pt end-pt
             #:width width
             rest))))

(define-method (make-pad-x length width . rest)
  (apply make-pad length width *along-x* rest))

(define-method (make-pad-y length width . rest)
  (apply make-pad length width *along-y* rest))

(define-method (convert-mm->centimils (pad <pad>))
  (let ((pad (next-method)))
    (slot-mod! pad 'mask-extent (convert-mm->centimils <>))
    (slot-mod! pad 'clearance-extent (convert-mm->centimils <>))
    pad))

(define-method (mask-length (pad <pad>))
  (+ (stroke-length pad)
     (* 2 (slot-ref pad 'mask-extent))))

(define-method (mask-width (pad <pad>))
  (+ (stroke-width pad)
     (* 2 (slot-ref pad 'mask-extent))))

(define-method (clearance-length (pad <pad>))
  (+ (stroke-length pad)
     (* 2 (slot-ref pad 'clearance-extent))))

(define-method (clearance-width (pad <pad>))
  (+ (stroke-width pad)
     (* 2 (slot-ref pad 'clearance-extent))))

(define-method (output (pad <pad>) out level)
  (let ((pad (if (>= (stroke-length pad) (stroke-width pad))
                 pad
                 (swap-axis pad))))
    (with-slots pad ((start-pt 'start-pt)
                     (end-pt 'end-pt)
                     (width 'width)
                     (clearance-extent 'clearance-extent)
                     (mask-extent 'mask-extent)
                     (name 'name)
                     (number 'number)
                     (flags 'flags))
      (let ((direction (unit-vector start-pt end-pt)))
        (let ((start-pt (move-along start-pt (/ width 2) direction))
              (end-pt (move-along end-pt (/ width 2) (- direction)))
              (mask (+ width (* 2 mask-extent)))
              (clearance (* 2 clearance-extent)))
          (format out
                  "~APad[~A ~A ~A ~A ~A ~S ~S ~S]"
                  (indent level)
                  (output start-pt #f)
                  (output end-pt #f)
                  (output (to-exact width) #f)
                  (output (to-exact clearance) #f)
                  (output (to-exact mask) #f)
                  (output name #f)
                  (output number #f)
                  (output flags #f)))))))

;;;
;;; element-line
;;;

(define-class <element-line> (<stroke>))

(define-method (make-element-line (start-pt <pt>) (end-pt <pt>) width)
  (make <element-line> #:start-pt start-pt #:end-pt end-pt #:width width))

;;; Make an element-line centered on the origin of the given LENGTH
;;; and WIDTH, with the length along the DIRECTION-UNIT-VECT
;;; direction.
(define-method (make-element-line length width (direction-unit-vect <pt>))
  (let ((half-length (/ length 2)))
    (let* ((end-pt (make-pt (* (x direction-unit-vect) half-length)
                            (* (y direction-unit-vect) half-length)))
           (start-pt (move-along end-pt (- length) direction-unit-vect)))
      (make-element-line start-pt end-pt width))))

(define-method (make-element-line-x length width)
  (make-element-line length width *along-x*))

(define-method (make-element-line-y length width)
  (make-element-line length width *along-y*))

(define-method (output (line <element-line>) out level)
  (with-slots line ((start-pt 'start-pt)
                    (end-pt 'end-pt)
                    (width 'width))
    (format out
            "~AElementLine[~A ~A ~A]"
            (indent level)
            (output start-pt #f)
            (output end-pt #f)
            (output (to-exact width) #f))))

(define-class <centroidal> ()
  (center-pt #:accessor center-pt
             #:init-keyword #:center-pt
             #:init-value (make-pt 0 0)))

(define-slots-ops ((move (obj <centroidal>) dx dy)
                   (move-along (obj <centroidal>) distance
                               (direction-unit-vect <pt>))
                   (rotate (obj <centroidal>) degrees))
  (center-pt))

(define-method (center (obj <centroidal>))
  (slot-ref obj 'center-pt))

(define-class <element-arc> (<centroidal>)
  (width #:init-keyword #:width
         #:init-value 0)
  (height #:init-keyword #:height
          #:init-value 0)
  (start-angle #:init-keyword #:start-angle
               #:init-value 0)
  (delta-angle #:init-keyword #:delta-angle
               #:init-value 0)
  (thickness #:accessor thickness
             #:init-keyword #:thickness
             #:init-value 0))

;; Make arc at origin
(define (make-element-arc radius-x radius-y start-angle delta-angle thickness)
  (make <element-arc>
    #:width radius-x
    #:height radius-y
    #:start-angle start-angle
    #:delta-angle delta-angle
    #:thickness thickness))

;; Make circle at origin
(define (make-element-circle radius thickness)
  (let ((diameter (* 2 radius)))
    (make-element-arc radius radius 0 360 thickness)))

(define-slots-op (convert-mm->centimils (arc <element-arc>))
  (center-pt
   width
   height
   thickness))

;;; For <element-arc>, mirroring and rotating need special handling.
(define-method (rotate (arc <element-arc>) degrees)
  (let ((circular? (= (slot-ref arc 'width) (slot-ref arc 'height))))
    ;; Non-circular arcs can only be rotated by multiples of 90 degrees
    (unless (or circular? (zero? (modulo degrees 90)))
      (error (string-append
              "You can only rotate non-circular element-arc by angles "
              "that are a multiple of 90 degrees")
             degrees arc))
    (let-deep-clone ((arc arc))
      (slot-mod! arc 'center-pt (rotate <> degrees))
      (slot-mod! arc 'start-angle (+ <> degrees))
      (unless (or circular? (zero? (modulo degrees 180)))
        ;; Swap width and height.
        (with-slots arc ((width 'width)
                         (height 'height))
          (slot-set! arc 'width height)
          (slot-set! arc 'height width)))
      arc)))

(define-method (mirror-x (arc <element-arc>))
  (let-deep-clone ((arc arc))
    (slot-mod! arc 'start-angle (- <>))
    (slot-mod! arc 'delta-angle (- <>))
    arc))

(define-method (mirror-y (arc <element-arc>))
  (let-deep-clone ((arc arc))
    (slot-mod! arc 'start-angle (- 180 <>))
    (slot-mod! arc 'delta-angle (- <>))
    arc))

(define-method (output (arc <element-arc>) out level)
  (format out
          "~AElementArc[~A ~A ~A ~A ~A ~A]"
          (indent level)
          (output (center-pt arc) #f)
          (output (to-exact (slot-ref arc 'width)) #f)
          (output (to-exact (slot-ref arc 'height)) #f)
          (output (slot-ref arc 'start-angle) #f)
          (output (slot-ref arc 'delta-angle) #f)
          (output (to-exact (thickness arc)) #f)))

(define-class <pin> (<flagable> <nameable> <numberable> <centroidal>)
  (thickness #:accessor thickness
             #:init-keyword #:thickness
             #:init-value 0)
  (mask-extent #:init-keyword #:mask-extent
               #:init-value 0)
  (clearance-extent #:init-keyword #:clearance-extent
                    #:init-value 0)
  (drill #:init-keyword #:drill #:init-value 0))

(define-slots-op (convert-mm->centimils (obj <pin>))
  (center-pt
   thickness
   mask-extent
   clearance-extent
   drill))

(define-method (output (obj <pin>) out level)
  (with-slots obj ((center-pt 'center-pt)
                   (thickness 'thickness) ; outer diameter
                   (mask-extent 'mask-extent)
                   (clearance-extent 'clearance-extent)
                   (drill 'drill)       ; diameter
                   (name 'name)
                   (number 'number)
                   (flags 'flags))
    (let ((mask (+ thickness (* 2 mask-extent))))
      (format out
              "~APin[~A ~A ~A ~A ~A ~S ~S ~S]"
              (indent level)
              (output center-pt #f)
              (output (to-exact thickness) #f)
              (output (to-exact clearance-extent) #f)
              (output (to-exact mask) #f)
              (output (to-exact drill) #f)
              (output name #f)
              (output number #f)
              (output flags #f)))))

(define-class <via> (<flagable> <nameable> <centroidal>)
  (thickness #:accessor thickness
             #:init-keyword #:thickness
             #:init-value 0)
  (mask-extent #:init-keyword #:mask-extent
               #:init-value 0)
  (clearance-extent #:init-keyword #:clearance-extent
                    #:init-value 0)
  (drill #:init-keyword #:drill #:init-value 0))

(define-slots-op (convert-mm->centimils (obj <via>))
  (center-pt
   thickness
   mask-extent
   clearance-extent
   drill))

(define-method (output (obj <via>) out level)
  (with-slots obj ((center-pt 'center-pt)
                   (thickness 'thickness) ; outer diameter
                   (mask-extent 'mask-extent)
                   (clearance-extent 'clearance-extent)
                   (drill 'drill)       ; diameter
                   (name 'name)
                   (flags 'flags))
    (let ((mask (+ thickness (* 2 mask-extent))))
      (format out
              "~AVia[~A ~A ~A ~A ~A ~S ~S]"
              (indent level)
              (output center-pt #f)
              (output (to-exact thickness) #f)
              (output (to-exact clearance-extent) #f)
              (output (to-exact mask) #f)
              (output (to-exact drill) #f)
              (output name #f)
              (output flags #f)))))

(define-class <element> ()
  (flags #:init-keyword #:flags #:init-value "")
  (description #:init-keyword #:description #:init-value "")
  (name #:init-keyword #:name #:init-value "")
  (value #:init-keyword #:value #:init-value "")
  (mark-pt #:init-keyword #:mark-pt #:init-value (make-pt 0 0))
  (text-pt #:init-keyword #:text-pt #:init-value (make-pt 0 0))
  (text-direction #:init-keyword #:text-direction #:init-value 0)
  (text-scale #:init-keyword #:text-scale #:init-value 100)
  (text-flags #:init-keyword #:text-flags #:init-value "")
  (contents #:init-keyword #:contents #:init-value '()))

(define-slots-ops ((move (obj <element>) dx dy)
                   (move-along (obj <element>) distance
                               (direction-unit-vect <pt>))
                   (rotate (obj <element>) degrees)
                   (mirror-x (obj <element>))
                   (mirror-y (obj <element>))
                   (convert-mm->centimils (obj <element>)))
  (mark-pt
   text-pt
   contents))

(define-method (output (elem <element>) out level)
  (format out
          "~AElement[~S ~S ~S ~S ~A ~A ~A ~A ~S] (~%~A~%)~%"
          (indent level)
          (output (slot-ref elem 'flags) #f)
          (output (slot-ref elem 'description) #f)
          (output (slot-ref elem 'name) #f)
          (output (slot-ref elem 'value) #f)
          (output (slot-ref elem 'mark-pt) #f)
          (output (slot-ref elem 'text-pt) #f)
          (output (slot-ref elem 'text-direction) #f)
          (output (slot-ref elem 'text-scale) #f)
          (output (slot-ref elem 'text-flags) #f)
          (output (slot-ref elem 'contents) #f (+ level 1))))

(define-method (make-outline dx dy thickness)
  (let ((line-x (make-element-line-x dx thickness))
        (line-y (make-element-line-y dy thickness)))
    (combine (move-dy line-x (/ dy 2))
             (move-dy line-x (- (/ dy 2)))
             (move-dx line-y (/ dx 2))
             (move-dx line-y (- (/ dx 2))))))

(define (remove-flag flags flag)
  (let ((flags (delete-duplicates (string-split flags #\space) string=?)))
    (string-join
     (delete flag (delete "" flags string=?) string=?)
     " ")))

(define (remove-flags flags flag-lst)
  (fold-right (lambda (f result)
                (remove-flag result f))
              flags
              flag-lst))

(define (add-flag flags flag)
  (let ((flags (delete-duplicates (string-split flags #\space) string=?)))
    (string-join
     (append (delete "" flags string=?)
             (list flag))
     " ")))

(define (replace-flag flags desired replace-lst)
  (add-flag (remove-flags flags replace-lst)
            desired))

(define *shape-flags* '("square" "octagon"))

(define-method (shape-round (obj <flagable>))
  (let-deep-clone ((obj obj))
    (slot-mod! obj 'flags
               (remove-flags <> *shape-flags*))
    obj))

(define-method (shape-round (objs <list>))
  (map shape-round objs))

(define-method (shape-round (objs <list>) (n <integer>))
  (modify-element shape-round objs n))

(define (shape-round-first objs)
  (shape-round objs 0))

(define (shape-round-last objs)
  (shape-round objs (- (length objs) 1)))

(define-method (shape-square (obj <flagable>))
  (let-deep-clone ((obj obj))
    (slot-mod! obj 'flags
               (replace-flag <> "square" *shape-flags*))
    obj))

(define-method (shape-square (objs <list>))
  (map shape-square objs))

(define-method (shape-square (objs <list>) (n <integer>))
  (modify-element shape-square objs n))

(define (shape-square-first objs)
  (shape-square objs 0))

(define (shape-square-last objs)
  (shape-square objs (- (length objs) 1)))

(define (to-string obj)
  (format #f "~A" obj))

;;; Sets a nameable object's name and number
(define-method (name (nameable <nameable>) name)
  (let-deep-clone ((nameable nameable))
    (let ((name (to-string name)))
      (slot-set! nameable 'name name)
      nameable)))

(define-method (number (numberable <numberable>) number)
  (let-deep-clone ((numberable numberable))
    (let ((number (to-string number)))
      (slot-set! numberable 'number number)
      numberable)))

(define-method (name* name-number-able name)
  (let-deep-clone ((name-number-able name-number-able))
    (let ((name (to-string name)))
      (slot-set! name-number-able 'name name)
      (slot-set! name-number-able 'number name)
      name-number-able)))

(define-method (enumerate nameables (start <number>))
  (enumerate nameables (iota (length nameables) start)))

(define-method (enumerate* nameables (start <number>))
  (enumerate* nameables (iota (length nameables) start)))

(define-method (enumerate-base nameables (lst <list>) name-fn)
  (let loop ((nameables nameables)
             (lst lst)
             (acc '()))
    (if (or (null? nameables) (null? lst))
        (reverse! acc)
        (loop (cdr nameables)
              (cdr lst)
              (cons (name-fn (car nameables) (car lst)) acc)))))

(define-method (enumerate nameables (lst <list>))
  (enumerate-base nameables lst name))

(define-method (enumerate* nameables (lst <list>))
  (enumerate-base nameables lst name*))

;;; Make N duplicates of OBJ starting at START-PT at DX and DY
;;; intervals.
(define-method (duplicate obj n (start-pt <pt>) dx dy)
  (append-map (lambda (i)
                (let ((p (place obj (+ start-pt (make-pt (* i dx) (* i dy))))))
                  (if (list? p)
                      (apply combine p)
                      (list p))))
              (iota n)))

;;; Make N duplicates where the center of the row is the center of the
;;; original object.
(define (row obj n dx dy)
  (let ((offset-factor (/ (- n 1) 2)))
    (duplicate obj n
               (move (center obj)
                     (- (* offset-factor dx))
                     (- (* offset-factor dy)))
               dx dy)))

(define-method (make-polar-outline-top dx dy
                                       bar-distance
                                       thickness)
  (let ((bar-y-offset (- (/ dy 2) bar-distance))
        (outline (make-outline dx dy thickness)))
    (let ((bar (move-along-y (make-element-line-x dx thickness)
                             bar-y-offset)))
      (combine outline bar))))

(define-rotate-methods make-polar-outline-top
  ((make-polar-outline-right -90)
   (make-polar-outline-bottom 180)
   (make-polar-outline-left 90)))

(define (make-scant-outline dx dy thickness)
  (let ((line-y (make-element-line-y dy thickness)))
    (let ((left (move-along-x line-y (- (/ dx 2))))
          (right (move-along-x line-y (/ dx 2))))
      (combine left right))))

(define-method (make-scant-polar-outline-top dx dy
                                             bar-distance
                                             thickness)
  (let ((scant-outline (make-scant-outline dx dy thickness)))
    (let* ((bar-y-offset (- (- (/ dy 2) bar-distance)))
           (bar (move-along-y (make-element-line-x dx thickness)
                              bar-y-offset)))
      (combine scant-outline bar))))

(define-rotate-methods make-scant-polar-outline-top
  ((make-scant-polar-outline-right -90)
   (make-scant-polar-outline-bottom 180)
   (make-scant-polar-outline-left 90)))

(define-method (make-notched-outline-ul dx dy notch-length thickness)
  (let ((line-x (move-along-y (make-element-line-x dx thickness)
                              (/ dy 2)))
        (line-y (move-along-x (make-element-line-y dy thickness)
                              (/ dx 2)))
        (projected-notch-length (* (sqrt 2) notch-length)))
    (let ((notch-line-x (move
                         (make-element-line-x (- dx projected-notch-length)
                                              thickness)
                         (/ projected-notch-length 2)
                         (- (/ dy 2))))
          (notch-line-y (move
                         (make-element-line-y (- dy projected-notch-length)
                                              thickness)
                         (- (/ dx 2))
                         (/ projected-notch-length 2))))
      (let ((notch-line (make-element-line (slot-ref notch-line-x 'start-pt)
                                           (slot-ref notch-line-y 'start-pt)
                                           thickness)))
        (combine line-x
                 line-y
                 notch-line-x
                 notch-line-y
                 notch-line)))))

(define-rotate-methods make-notched-outline-ul
  ((make-notched-outline-ur -90)
   (make-notched-outline-lr 180)
   (make-notched-outline-ll 90)))

(define-method (make-notched-outline-top dx dy
                                         notch-length thickness)
  (let ((line-x (move-along-y (make-element-line-x dx thickness)
                              (/ dy 2)))
        (line-y (make-element-line-y dy thickness))
        (notch-line-x (make-element-line-x (- (/ dx 2) (/ notch-length 2))
                                         thickness)))
    (combine line-x
             (move-along-x line-y (- (/ dx 2)))
             (move-along-x line-y (/ dx 2))
             (move notch-line-x (- (/ (+ dx notch-length) 4)) (- (/ dy 2)))
             (move notch-line-x (/ (+ dx notch-length) 4) (- (/ dy 2)))
             (make <element-arc>
               #:center-pt (make-pt 0 (- (/ dy 2)))
               #:width (/ notch-length 2)
               #:height (/ notch-length 2)
               #:start-angle 0
               #:delta-angle 180
               #:thickness thickness))))

(define-rotate-methods make-notched-outline-top
  ((make-notched-outline-right -90)
   (make-notched-outline-bottom 180)
   (make-notched-outline-left 90)))

(define (output-to-file obj file-name)
  (call-with-output-file file-name
    (lambda (out)
      (output obj out))))
